function [ V, energy ] = calculate_p_voltages( W, p, label_idxs, labels )
%CALCULATE_P_VOLTAGES Computes the p-voltage labels for a graph.
%
%  This solves the optimization problem described in Equation 3 of 
%  Bridle and Zhu, 2013, which minimizes a p-electric energy functional over
%  a graph. You need to have the Matlab Optimization Toolbox installed in 
%  order to use this function.
%
%  Input:
%    W: the weight matrix of the graph
%
%    p: the main parameter of the algorithm. Determines how sparsely current
%       flows in the network. p=2 corresponds to a physical electric network.  
%
%    label_idxs: the list of indexes of labeled nodes. For example, [1 3 4]
%       means that nodes 1, 3, 4 have fixed label.
%
%    labels: the list of labels of the labeled nodes. For example, [1 0 1]
%       means that the labeled nodes have labels 1, 0, and 1 respectively.
%
%
%  Output:
%    V: a vector containing the calculated p-voltages for all nodes in the
%      graph.
%    energy: the minimized total energy of the network
%
%  Author: Nick Bridle, nbridle@cs.wisc.edu. January 2013.
%
%  Citation:
%    Bridle, Nick and Xiaojin Zhu. 
%    p-voltages: Laplacian Regularization for Semi-Supervised Learning on 
%      High-Dimensional Data.
%    In the Eleventh Workshop on Mining and Learning with Graphs, MLG 2013

q = p/(p-1);
num_labels = length(labels);
u = length(W(:,1)) - num_labels;

v0 = rand(u,1);
[r,c] = find(W>0);
idx = find(W>0);

W_nz = W(idx).^(q-1);

unlabeled = setdiff(1:(num_labels+u), label_idxs);
func = @(f)compute_norm(f, q, W_nz, idx, r, c, unlabeled, label_idxs, labels);

% Solve the optimization problem
% We use the large-scale algorithm which uses the computed Hessian of the
% function.
% Set 'Display' to 'iter-detailed' if you would like to see details about
% the optimization as it progresses.
options = optimset('LargeScale', 'on', 'Display', 'off', 'Hessian', 'on', 'GradObj', 'on', 'TolFun',1e-18, 'TolX',1e-12);
[solution, energy] = fminunc(func, v0, options);

% Insert the fixed labels in the correct place
V = zeros(length(solution)+num_labels,1)+NaN;
V(label_idxs) = labels;
V(isnan(V)) = solution;
