function [ D, W ] = create_knn_graph( X, degrees, weight_func, k )
%CREATE_KNN_GRAPH Creates an unweighted KNN graph from a set of datapoints
%   Computes the degree and weight matrices of a KNN graph given the
%   datapoints in space. 
%   X is the matrix of data points
%   degrees is the number of fixed neighbors for each labeled point (leave
%   empty if you do not want the degrees of any points changed)
%   weight_func is a function used to determine the weight between two
%   points based on their distance. For instance, a Gaussian kernel
%   k is the number of neighbors each point should have. Set to [] if you
%   want log(n) neighbors.
%   D is the output degree matrix
%   W is the output weight matrix

% Construct the Laplacian matrix
% We separately compute the degree and weight matrices; L = D-W
n = size(X,1);
if isempty(k)
    k = ceil(log(n))+1
end
D = zeros(n);
W = zeros(n);
for row = 1:n
    dists = zeros(1,n);
    % Find the k nearest neighbors of each point
    for col = 1:n        
        % Compute the Euclidean distance between two vectors
        dists(col) = norm(X(row,:) - X(col,:));
    end
    
    [values,indices] = sort(dists);
    
    % For each nearest neighbor, add an edge with unit weight if
    % it doesn't already exist
    for i = 2:k+1
        w = weight_func(values(i));
        sel = indices(i);
        if W(row,sel)==0 && W(sel,row)==0 
             D(row,row) = D(row,row) + w;
             D(sel,sel) = D(sel,sel) + w;
             W(row,sel) = W(row,sel) + w;
             W(sel,row) = W(sel,row) + w;
        end
    end
end

if(isempty(degrees))
   return 
end

% Ensure that labeled points are only connected to the number of points
% indicated in the 'degrees' array
numSources = length(degrees);
for row = 1:numSources
    connected = [];
    
    for col = 1:n
       if row == col
           continue;
       end
       
       if W(row,col) > 0
          connected = [connected col];
       end
    end
   
    if length(connected) < degrees(row)
       message = ['Error: labeled point ' num2str(row) ' has only '...
           num2str(length(connected)) ' neighbors, but ' num2str(degrees(row)) ...
           ' were requested.'];
       disp(message);
       return;
    end
    
    % The node has the expected number of neighbors
    if length(connected) == degrees(row)
        continue;
    end
    
    % The node has too many neighbors. Remove ones that are furthest away
    dists = [];
    for col = connected
        dist = sum((X(row,:)-X(col,:)).^2);
        dists = [dists dist];
    end
    
    [values, indices] = sort(dists);
    for i = degrees(row)+1:length(connected)
       col = connected(indices(i));
       w = W(row,col);
       W(row,col) = 0;
       W(col,row) = 0;
       D(row,row) = D(row,row)-w;
       D(col,col) = D(col,col)-w;
    end
end
